'''Generic functions related to working with files or the file system.'''

# pylint: disable=invalid-name
# pylint: disable=too-many-arguments
# pylint: disable=too-many-boolean-expressions

import asyncio
import os
import re
import time
import zipfile
from asyncio import sleep
from concurrent.futures import as_completed
from concurrent.futures import ThreadPoolExecutor
from datetime import datetime
from datetime import timezone
from datetime import tzinfo
from pathlib import Path
from typing import Any
from typing import Dict
from typing import List

import mandrill
from smb.base import SharedFile
from smb.SMBConnection import SMBConnection
from smb.smb_structs import OperationFailure

DIRECTORY = Path(__file__).parent.absolute()


def async_wrapper(func):
    '''Decorator to run functions using async. Found this handy to use with DAG tasks.'''

    def wrapper(*args, **kwargs):
        '''Decorator wrapper.'''

        loop = asyncio.get_event_loop()
        return loop.run_until_complete(func(*args, **kwargs))

    return wrapper


def async_wrapper_using_new_loop(func):
    '''Decorator to run functions using async. Found this handy to use with DAG tasks.'''

    def wrapper(*args, **kwargs):
        '''Decorator wrapper.'''

        return asyncio.run(func(*args, **kwargs))

    return wrapper


async def async_process_manager(
        function: asyncio.coroutine,
        list_of_kwargs: List[Dict[str, Any]],
        chunk_size: int,
        use_threads=True) -> List[Any]:
    '''Process manager to run fixed number of functions, usually the same function expressed as
    coroutines in an array.  Use case is sending many http requests or iterating files.'''

    results = []
    if use_threads:
        with ThreadPoolExecutor(max_workers=chunk_size) as exe:
            futures = [exe.submit(function, **items) for items in list_of_kwargs]
            for future in as_completed(futures):
                results.append(future.result())
    else:
        for num in range(0, len(list_of_kwargs), chunk_size):
            tasks = [function(**items) for items in list_of_kwargs[num:num+chunk_size]]
            results.extend(await asyncio.gather(*tasks))

    return results


async def unzip_file(filepath: str, directory: str) -> List[str]:
    '''Unzip supplied filepath in the supplied directory returning list of filenames.'''

    zipped = zipfile.ZipFile(filepath)

    # exclude __MACOSX directory that could be added when creating zip on macs
    filenames = [i for i in zipped.namelist() if '__MACOSX' not in i]

    zipped.extractall(directory)
    zipped.close()

    return filenames


async def unzip_file_get_filepaths(
        filepath: str,
        directory: str,
        include_extensions: List[str] = None,
        exclude_extensions: List[str] = None) -> List[str]:
    '''Get all the filepaths after unzipping supplied filepath in the supplied directory.
    If the zipfile contains zipfiles, those files will also be unzipped. If include_extensions
    is supplied then add those file types to the result. If exclude_extensions is supplied
    then skip adding those filepaths to the result.'''

    paths = []
    zipfile_filepaths = [filepath]
    while zipfile_filepaths:

        new_zipfile_filepaths = []
        for path in zipfile_filepaths:

            for filename in await unzip_file(filepath=path, directory=directory):
                filepath = os.path.join(directory, filename)
                suffix = Path(filepath).suffix.lower()[1:]
                if suffix in 'zip':
                    new_zipfile_filepaths.append(filepath)
                elif include_extensions:
                    if suffix in include_extensions:
                        paths.append(filepath)
                elif exclude_extensions:
                    if suffix not in exclude_extensions:
                        paths.append(filepath)
                else:
                    paths.append(filepath)

        zipfile_filepaths = new_zipfile_filepaths

    return paths


async def get_current_datetime_from_timestamp(
        dt_format: str = '%Y-%m-%d %H_%M_%S.%f',
        time_zone: tzinfo = timezone.utc) -> str:
    '''Get the datetime from the timestamp in the format and timezone desired.'''

    return datetime.fromtimestamp(time.time(), time_zone).strftime(dt_format)


async def send_emails_via_mandrill(
        mandrill_api_key: str,
        emails: List[str],
        subject: str,
        global_merge_vars: List[Dict[str, Any]],
        template_name: str,
        template_content: List[Dict[str, Any]] = None
        ) -> Any:
    '''Send emails via Mailchimp mandrill API.'''

    message = {
        'to': [{'email': email} for email in emails],
        'subject': subject,
        'merge_language': 'handlebars',
        'global_merge_vars': global_merge_vars
    }

    return mandrill.Mandrill(mandrill_api_key).messages.send_template(
        template_name=template_name,
        template_content=template_content,
        message=message
    )


async def establish_ftp_connection(
        user: str,
        pwd: str,
        name: str,
        server: str,
        dns: str,
        port: int = 139,
        use_ntlm_v2: bool = True,
        is_direct_tcp: bool = False) -> SMBConnection:
    '''Establish FTP connection'''

    conn = SMBConnection(
        username=user,
        password=pwd,
        my_name=name,
        remote_name=server,
        use_ntlm_v2=use_ntlm_v2,
        is_direct_tcp=is_direct_tcp
    )
    conn.connect(ip=dns, port=port)
    return conn


async def list_ftp_objects(
        conn: SMBConnection,
        service_name: str,
        ftp_path: str,
        exclude_directories: bool = False,
        exclude_files: bool = False,
        regex_pattern: str = None) -> List[SharedFile]:
    '''List all files and directories in an FTP directory.'''

    results = []
    for item in conn.listPath(service_name, ftp_path):
        is_directory = item.isDirectory
        if item.filename == '.' or item.filename == '..' or \
                (exclude_directories and is_directory) or (exclude_files and not is_directory):
            continue
        if regex_pattern is None or re.search(regex_pattern, item.filename) is not None:
            results.append(item)

    return results


async def delete_ftp_file(conn: SMBConnection, service_name: str, ftp_path: str) -> bool:
    '''Remove a file from FTP and verify deletion.'''

    status = False
    conn.deleteFiles(service_name, ftp_path)
    try:
        conn.getAttributes(service_name, ftp_path)
    except OperationFailure:
        status = True

    return status


async def write_file_to_ftp(
        conn: SMBConnection,
        service_name: str,
        ftp_path: str,
        local_filepath) -> SharedFile:
    '''Write file to FTP creating missing FTP directories if necessary.'''

    # steps to create missing directories
    path = ''
    for directory in os.path.dirname(ftp_path).split(os.sep):
        folders = {i.filename for i in conn.listPath(service_name=service_name, path=path)}
        if directory not in folders:
            conn.createDirectory(service_name=service_name, path=f'{path}/{directory}')
            await sleep(1)
        path = directory if not path else f'{path}/{directory}'

    # write local file to FTP server
    with open(local_filepath, 'rb') as file_obj:
        conn.storeFile(service_name=service_name, path=ftp_path, file_obj=file_obj, timeout=300)
        await sleep(1)

    # return file attributes
    return conn.getAttributes(service_name=service_name, path=ftp_path)



'''aioradio redis cache script.'''

# pylint: disable=c-extension-no-member
# pylint: disable=too-many-instance-attributes

import asyncio
import hashlib
from dataclasses import dataclass
from dataclasses import field
from typing import Any
from typing import Dict
from typing import List

import aioredis
from fakeredis.aioredis import create_redis_pool as fake_redis_pool
import orjson

HASH_ALGO_MAP = {
    'SHA1': hashlib.sha1,
    'SHA224': hashlib.sha224,
    'SHA256': hashlib.sha256,
    'SHA384': hashlib.sha384,
    'SHA512': hashlib.sha512,
    'SHA3_224': hashlib.sha3_224,
    'SHA3_256': hashlib.sha3_256,
    'SHA3_384': hashlib.sha3_384,
    'SHA3_512': hashlib.sha3_512
}


@dataclass
class Redis:
    '''class dealing with aioredis functions.'''

    config: Dict[str, Any] = field(default_factory=dict)
    pool: aioredis.Redis = field(init=False, repr=False)
    pool_task: asyncio.coroutine = None

    # Set the redis pool min and max connections size
    pool_minsize: int = 5
    pool_maxsize: int = 10

    # Cache expiration in seconds
    expire: int = 60

    # If the key contains sensitive info than you can hash the cache key using various hash algorithms
    use_hashkey: bool = False
    hash_algorithm: str = 'SHA3_256'

    # If you want to pass in an object and let this class convert to json or
    # retrieve value letting this class convert from json set use_json = True.
    use_json: bool = True

    # used exclusively for pytest
    fakeredis: bool = False

    def __post_init__(self) -> None:
        '''Post constructor'''

        if self.fakeredis:
            self.pool = asyncio.get_event_loop().run_until_complete(fake_redis_pool())
        else:
            primary_endpoint = f'redis://{self.config["redis_primary_endpoint"]}'
            loop = asyncio.get_event_loop()
            if loop and loop.is_running():
                self.pool_task = loop.create_task(
                    aioredis.create_redis_pool(primary_endpoint, minsize=self.pool_minsize, maxsize=self.pool_maxsize))
            else:
                self.pool = loop.run_until_complete(
                    aioredis.create_redis_pool(primary_endpoint, minsize=self.pool_minsize, maxsize=self.pool_maxsize))

    def __del__(self) -> None:
        '''Teardown function'''

        self.pool.close()

    async def get_one_item(self, cache_key: str, use_json: bool=None) -> str:
        '''Check if an item is cached in redis.'''

        if use_json is None:
            use_json = self.use_json

        value = await self.pool.get(cache_key)

        if value is not None and use_json:
            value = orjson.loads(value)

        return value

    async def get_many_items(self, items: List[str], use_json: bool=None) -> List[str]:
        '''Check if many items are cached in redis.'''

        if use_json is None:
            use_json = self.use_json

        values = await self.pool.mget(*items)

        if use_json:
            values = [orjson.loads(val) if val is not None else None for val in values]

        return values

    async def set_one_item(self, cache_key: str, cache_value: str, expire: int=None, use_json: bool=None) -> None:
        '''Set one key-value pair in redis.'''

        if expire is None:
            expire = self.expire

        if use_json is None:
            use_json = self.use_json

        if use_json:
            cache_value = orjson.dumps(cache_value)

        await self.pool.set(cache_key, cache_value, expire=expire)

    async def delete_one_item(self, cache_key: str) -> None:
        '''Delete key from redis.'''

        return await self.pool.delete(cache_key)

    async def build_cache_key(self, payload: Dict[str, Any], separator='|', use_hashkey: bool=None) -> str:
        '''If you'd like to build a cache key from a dictionary object this is the function for you.
        This funciton will concatenate and normalize key-values from an unnested dict, taking
        care of sorting the keys and each of their values (if a list).'''

        if use_hashkey is None:
            use_hashkey = self.use_hashkey

        keys = sorted(payload.keys())
        concat_key = separator.join([f'{k}={orjson.dumps(payload[k], option=orjson.OPT_SORT_KEYS).decode()}'.replace('"', '')
                                    for k in keys if payload[k] != [] and payload[k] is not None])

        if use_hashkey:
            concat_key = HASH_ALGO_MAP[self.hash_algorithm](concat_key.encode()).hexdigest()

        return concat_key



'''Generic async AWS functions for DynamoDB.'''

# pylint: disable=too-many-arguments

from typing import Any
from typing import Dict
from typing import List

from aioradio.aws.utils import AwsServiceManager

AWS_SERVICE = AwsServiceManager(service='dynamodb', regions=['us-east-1', 'us-east-2'], module='aioboto3')
DYNAMO = AWS_SERVICE.service_dict


@AWS_SERVICE.active
async def create_dynamo_table(
        table_name: str,
        region: str,
        attribute_definitions: List[Dict[str, str]],
        key_schema: List[Dict[str, str]],
        provisioned_throughput: Dict[str, int]) -> str:
    '''Create dynamo table. '''

    error = ''

    try:
        table = await DYNAMO[region]['client']['obj'].create_table(
            TableName=table_name,
            AttributeDefinitions=attribute_definitions,
            KeySchema=key_schema,
            ProvisionedThroughput=provisioned_throughput
        )
        await table.meta.client.get_waiter('table_exists').wait(TableName=table_name)
    except DYNAMO[region]['client']['obj'].exceptions.ResourceInUseException as err:
        error = err.response['Error']['Message']

    return error


@AWS_SERVICE.active
async def get_list_of_dynamo_tables(region: str) -> List[str]:
    '''Get list of Dynamo tables in a particular region.'''

    tables = []
    result = await DYNAMO[region]['client']['obj'].list_tables()
    tables = result['TableNames']

    return tables


@AWS_SERVICE.active
async def scan_dynamo(table_name: str, region: str, key: Any=None) -> List[Any]:
    '''Scan dynamo table using a filter_expression if supplied.'''

    result = []
    scan_kwargs = {'FilterExpression': key} if key is not None else {}

    while True:
        table = await DYNAMO[region]['resource']['obj'].Table(table_name)
        resp = await table.scan(**scan_kwargs)
        result.extend(resp['Items'])
        if 'LastEvaluatedKey' not in resp:
            break
        scan_kwargs['ExclusiveStartKey'] = resp['LastEvaluatedKey']

    return result


@AWS_SERVICE.active
async def put_item_in_dynamo(table_name: str, region: str, item: Dict[str, Any]) -> Dict[str, Any]:
    '''Put item in dynamo table.'''

    result = {}
    table = await DYNAMO[region]['resource']['obj'].Table(table_name)
    result = await table.put_item(Item=item)
    return result


@AWS_SERVICE.active
async def query_dynamo(table_name: str, region: str, key: Any):
    '''Query dynamo for with specific key_condition_expression.'''

    result = []
    query_kwargs = {'KeyConditionExpression': key}

    while True:
        table = await DYNAMO[region]['resource']['obj'].Table(table_name)
        resp = await table.query(**query_kwargs)
        result.extend(resp['Items'])
        if 'LastEvaluatedKey' not in resp:
            break
        query_kwargs['ExclusiveStartKey'] = resp['LastEvaluatedKey']

    return result


@AWS_SERVICE.active
async def update_item_in_dynamo(
        table_name: str,
        region: str,
        key: Dict[str, Any],
        update_expression: str,
        expression_attribute_names: str,
        expression_attribute_values: str='',
        condition_expression: str='',
        return_values: str='UPDATED_NEW') -> Dict[str, Any]:
    '''Update an item in Dynamo without overwriting the entire item.'''

    result = {}
    update_kwargs = {
        'Key': key,
        'UpdateExpression': update_expression,
        'ExpressionAttributeNames': expression_attribute_names,
        'ReturnValues': return_values
    }
    if expression_attribute_values:
        update_kwargs['ExpressionAttributeValues'] = expression_attribute_values
    if condition_expression:
        update_kwargs['ConditionExpression'] = condition_expression

    table = await DYNAMO[region]['resource']['obj'].Table(table_name)
    result = await table.update_item(**update_kwargs)

    return result



@AWS_SERVICE.active
async def batch_write_to_dynamo(table_name: str, region: str, items: List[Dict[str, Any]]) -> bool:
    '''Write batch of items to dynamo table.'''

    batch_writer_successful = False

    table = await DYNAMO[region]['resource']['obj'].Table(table_name)
    async with table.batch_writer() as db_writer:
        for item in items:
            await db_writer.put_item(Item=item)

        batch_writer_successful = True

    return batch_writer_successful


@AWS_SERVICE.active
async def batch_get_items_from_dynamo(
        table_name: str,
        region: str,
        items: List[Dict[str, Any]]) -> Dict[str, Any]:
    '''Get batch of items from dynamo.'''

    response = await DYNAMO[region]['resource']['obj'].batch_get_item(RequestItems={table_name: {'Keys': items}})

    return response
